"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts_lodash_1 = require("ts-lodash");
const args_1 = require("./args");
const flags_1 = require("./flags");
const validate_1 = require("./validate");
function parse(options) {
    const input = {
        args: (options.args || []).map(a => args_1.newArg(a)),
        argv: options.argv || process.argv.slice(2),
        flags: Object.assign({ color: flags_1.defaultFlags.color }, options.flags || {}),
        strict: options.strict !== false,
    };
    const parser = new Parser(input);
    const output = parser.parse();
    validate_1.validate(input, output);
    return output;
}
exports.parse = parse;
class Parser {
    constructor(input) {
        this.input = input;
        this.raw = [];
        this.argv = input.argv.slice(0);
        this._setNames();
        this.booleanFlags = ts_lodash_1.default.pickBy(input.flags, (f) => f.type === 'boolean');
    }
    parse() {
        const findLongFlag = (arg) => {
            const name = arg.slice(2);
            if (this.input.flags[name]) {
                return name;
            }
            if (arg.startsWith('--no-')) {
                const flag = this.booleanFlags[arg.slice(5)];
                if (flag && flag.allowNo)
                    return flag.name;
            }
        };
        const findShortFlag = (arg) => {
            return Object.keys(this.input.flags).find(k => this.input.flags[k].char === arg[1]);
        };
        const parseFlag = (arg) => {
            const long = arg.startsWith('--');
            const name = long ? findLongFlag(arg) : findShortFlag(arg);
            if (!name) {
                const i = arg.indexOf('=');
                if (i !== -1) {
                    const sliced = arg.slice(i + 1);
                    this.argv.unshift(sliced);
                    const equalsParsed = parseFlag(arg.slice(0, i));
                    if (!equalsParsed) {
                        this.argv.shift();
                    }
                    return equalsParsed;
                }
                return false;
            }
            const flag = this.input.flags[name];
            if (flag.type === 'option') {
                let input;
                if (long || arg.length < 3) {
                    input = this.argv.shift();
                }
                else {
                    input = arg.slice(arg[2] === '=' ? 3 : 2);
                }
                if (!input) {
                    throw new Error(`Flag --${name} expects a value`);
                }
                this.raw.push({ type: 'flag', flag: flag.name, input });
            }
            else {
                this.raw.push({ type: 'flag', flag: flag.name, input: arg });
                if (!long && arg.length > 2) {
                    this.argv.unshift(`-${arg.slice(2)}`);
                }
            }
            return true;
        };
        let parsingFlags = true;
        while (this.argv.length) {
            const input = this.argv.shift();
            if (parsingFlags && input.startsWith('-')) {
                if (input === '--') {
                    parsingFlags = false;
                    continue;
                }
                if (parseFlag(input)) {
                    continue;
                }
            }
            this.raw.push({ type: 'arg', input });
        }
        const argv = this._argv();
        return {
            args: this._args(argv),
            argv,
            flags: this._flags(),
            raw: this.raw,
        };
    }
    _args(argv) {
        const args = {};
        for (let i = 0; i < this.input.args.length; i++) {
            const arg = this.input.args[i];
            args[arg.name] = argv[i];
        }
        return args;
    }
    _flags() {
        const flags = {};
        for (const token of this._flagTokens) {
            const flag = this.input.flags[token.flag];
            if (!flag)
                throw new Error(`Unexpected flag ${token.flag}`);
            if (flag.type === 'boolean') {
                if (token.input === `--no-${flag.name}`) {
                    flags[token.flag] = false;
                }
                else {
                    flags[token.flag] = true;
                }
            }
            else {
                const value = flag.parse(token.input, {});
                if (flag.multiple) {
                    flags[token.flag] = flags[token.flag] || [];
                    flags[token.flag].push(value);
                }
                else {
                    flags[token.flag] = value;
                }
            }
        }
        for (const [k, flag] of Object.entries(this.input.flags)) {
            if (!flags[k]) {
                if (flag.type === 'option' && flag.default) {
                    flags[k] = flag.default;
                }
            }
        }
        return flags;
    }
    _argv() {
        const args = [];
        const tokens = this._argTokens;
        for (let i = 0; i < Math.max(this.input.args.length, tokens.length); i++) {
            const token = tokens[i];
            const arg = this.input.args[i];
            if (token) {
                args[i] = arg ? arg.parse(token.input) : token.input;
            }
            else {
                if (arg.default)
                    args[i] = arg.default;
            }
        }
        return args;
    }
    get _argTokens() {
        return this.raw.filter(o => o.type === 'arg');
    }
    get _flagTokens() {
        return this.raw.filter(o => o.type === 'flag');
    }
    _setNames() {
        for (const [name, flag] of Object.entries(this.input.flags)) {
            flag.name = name;
        }
    }
}
